
#include <defines.h>

#include "loadrunner.h"
#include "dirs.h"
#include "settings.h"
#include "qiconloader.h"
#include "about.h"

#ifdef _QTE_
#include <qpe/qpeapplication.h>
#include <qpe/qcopenvelope_qws.h>
#else
#include <qapplication.h>
#include <qfiledialog.h>
#endif

#include <qmenubar.h>
#include <qstatusbar.h>
#include <qtoolbar.h>
#include <qtoolbutton.h>
#include <qbuttongroup.h>
#include <qmessagebox.h>
#include <qfile.h>
#include <qdict.h>
#include <qaction.h>
#include <qlabel.h>
#include <qtextstream.h>
#include <qdir.h>

#include <stdlib.h>

// --------------------------------------------------------
// ---------------- LoadRunnerPrivate --------------------
// --------------------------------------------------------

class LoadRunnerPrivate
{
  public:
    LoadRunnerPrivate(LoadRunner * parent);
    virtual ~LoadRunnerPrivate();
    
    void createObject(KGrObj*, int, int);
    void changeObject(unsigned char, int, int);
    
    void deleteLevel();
    bool loadLevel(int);
    bool loadLevel(const QString & fileName);
  
    void showHidden();
    void pause();
    void stop();
    void nextLevel();
    void initMap();
    
    QDict<QAction> actions;
    QToolBar * toolBar;
    QButtonGroup * buttonGroup;
    QList<QToolButton> toolButtons;
    LoadRunnerWidget * view;
    LoadRunner::State state;
    QString levelName;
    QString saveFileName;
    
    FieldMap map;
    KGrHero * hero;
    EnemiesList enemies;

    int level;
    int lifes;
    int score;
    int nuggets;
    
  private:
  
    void initSearchMatrix();
    
    LoadRunner * parent_;
};

LoadRunnerPrivate::LoadRunnerPrivate(LoadRunner * parent)
  : toolBar(0L),
    buttonGroup(0L),
    view(0L),
    state( LoadRunner::Stop ),
    map(OBJECTS_X, OBJECTS_Y),
    hero(0L),
    level(0),
    lifes( LIFES_DEFAULT ),
    score(0),
    nuggets(0),
    parent_(parent)
{
  enemies.setAutoDelete(true);
 
  // build the view
  view = new LoadRunnerWidget(parent);
  view->resize((map.objectsX-2) * OBJECT_SIZE, (map.objectsY-2) * OBJECT_SIZE);
  view->setFixedSize((map.objectsX-2) * OBJECT_SIZE, (map.objectsY-2) * OBJECT_SIZE);
  view->setBackgroundColor( QObject::black );

  // init the map
  for(int i=0; i<map.objectsX; i++)
    for(int j=0; j<map.objectsY; j++)
      map.playField[i][j] = 0L;
    
  // build the default objects
  for(int i=0; i<map.objectsX-2; i++)
    for(int j=0; j<map.objectsY-2; j++)
    {
      map.playField[i+1][j+1] = new KGrFree(false, view);
      map.playField[i+1][j+1]->move(i*16, j*16);
    }
    
  // build the borders for X and for Y
  for (int i=0; i<map.objectsX; i++)
  {
    map.playField[i][0] = new KGrBeton();
    map.playField[i][map.objectsY-1] = new KGrBeton();
  }
  for (int j=0; j<map.objectsY; j++)
  {
    map.playField[0][j] = new KGrBeton();
    map.playField[map.objectsX-1][j] = new KGrBeton();
  }

  // check the map
  for(int i=0; i<map.objectsX; i++)
    for(int j=0; j<map.objectsY; j++)
    {
      if( !map.playField[i][j] )
      {
        state = LoadRunner::Error;
        qDebug("Cannot allocate the map.");
        break;
      }
    }
    
  hero = new KGrHero(0, 0);
  if( hero )
    hero->setPlayfield( &map );
  else
  {
    qDebug("Cannot allocate the hero.");
    state = LoadRunner::Error;
  }
}

LoadRunnerPrivate::~LoadRunnerPrivate()
{
  // delete the map
  for(int i=0; i<map.objectsX; i++)
    for(int j=0; j<map.objectsY; j++)
    {
      if( !map.playField[i][j] )
      {
        delete map.playField[i][j];
        map.playField[i][j] = 0L;
      }
    }
    
  if( hero )
  {
    delete hero;
    hero = 0;
  }
  
  enemies.clear();
  
  if( view )
  {
    delete view;
    view = 0L;
  }
  
  toolButtons.setAutoDelete(true);
  toolButtons.clear();
  
  if( buttonGroup )
  {
    delete buttonGroup;
    buttonGroup = 0L;
  }
  
  if( toolBar )
  {
    delete toolBar;
    toolBar = 0L;
  }
}

void LoadRunnerPrivate::initMap()
{
  // build the default objects
  for(int i=0; i<map.objectsX-2; i++)
    for(int j=0; j<map.objectsY-2; j++)
      changeObject(FREE, i+1, j+1);
  
  enemies.clear();
  hero->init(0, 0);
  hero->hide();
}

bool LoadRunnerPrivate::loadLevel(const QString & fileName)
{
//  qDebug("loadLevel(): File name = %s", fileName.latin1());
  
  int iRead=0;
  QFile openLevel(fileName);
  char buffer[ map.objectsX ];
  QString lineInfo;
    
  // initialize the game
  nuggets = 0;
  enemies.clear();
  state = LoadRunner::NewGame;
  
  if( !openLevel.open(IO_ReadOnly) )
  {
    QMessageBox::information( parent_, QObject::tr("Loading Level"),
                              QObject::tr("Error when opening the Level file.") );
    state = LoadRunner::Error;
    level = 0;
    return false;
  }
  
  // read the info lines
  while( -1 != openLevel.readLine(lineInfo, 256) &&
          "MAP" != lineInfo.left(3) )
  {
    // check the map version, etc.
    
    if( "NAME" == lineInfo.left(4) )
    {
      levelName = lineInfo.mid( lineInfo.find("=") + 1 );
      levelName = levelName.left( levelName.find('\n') );
    }
  }
  
  for(int j=1; j<map.objectsY-1; j++)
  {
    iRead = openLevel.readLine(buffer, map.objectsX);
    
    if( iRead < map.objectsX-1 )
    {
      qDebug("Wrong file format. %d from %d", iRead, map.objectsX);
      state = LoadRunner::Error;
      break;
    }
    
    for(int i=1; i<map.objectsX-1; i++)
      changeObject(buffer[i-1], i, j);
  }
    
  openLevel.close();
  
  hero->setEnemyList( &enemies );
  hero->setNuggets(nuggets);
  hero->showFigure();
  
  for(EnemiesListIterator it(enemies); it.current(); ++it)
  {
    it.current()->setEnemyList( &enemies );
    it.current()->showFigure();
  }
  
  initSearchMatrix();
  
  return true;
}

bool LoadRunnerPrivate::loadLevel(int levelNo)
{
  QString user, syst;

  user.setNum(levelNo); // INT -> QString
  user=user.rightJustify(3,'0'); // prepend zeros
  user.append(".lr"); // append goldrunnerlevel-sufix
  syst=user;
  user.prepend("levels/level");
  syst.prepend("levels/level");
  
  user = Dirs::locate(user);
  syst = Dirs::locate(syst);
       
  QFile userLevel(user);
  QFile systemLevel(syst);
  
  if( !userLevel.exists() )
  {
    if( !systemLevel.exists() )
    {
      QMessageBox::information(parent_, QObject::tr("Loading Level"),
        QObject::tr("This game level doesn't exist. \nPlease check your game installation.") );
      return false;
    }
    else 
    {
//    statusBar->changeItem("Systemlevel", ID_LEVELPLACE);
      return loadLevel( syst );
    } 
  }
  else
  {
//    statusBar->changeItem("Userlevel", ID_LEVELPLACE);
    return loadLevel( user );
  }

  return false;
}

void LoadRunnerPrivate::initSearchMatrix()
{
  int X(map.objectsX-1);
  int Y(map.objectsY-1);
  
  for (int j=1; j<X; j++)
  {
    for(int i=1; i<Y; i++)
    {
      if( LADDER == map.playField[j][i]->whatIam() )
        map.playField[j][i]->searchValue = CANWALKLEFT+CANWALKRIGHT+CANWALKUP+CANWALKDOWN;
      else
      {
        char field = map.playField[j][i+1]->whatIam();
        switch( field )
        {
          case BRICK:
          case FBRICK:
          case HOLE:
          case USEDHOLE:
          case BETON:
            map.playField[j][i]->searchValue = CANWALKLEFT+CANWALKRIGHT;
            break;
          case LADDER:
            map.playField[j][i]->searchValue = CANWALKLEFT+CANWALKRIGHT+CANWALKDOWN;
            break;
          default :
            if( POLE == map.playField[j][i]->whatIam() )
              map.playField[j][i]->searchValue = CANWALKLEFT+CANWALKRIGHT+CANWALKDOWN;
            else
              map.playField[j][i]->searchValue = CANWALKDOWN;
            break;
        }
      }
      
      if(map.playField[j][i-1]->blocker)
        map.playField[j][i]->searchValue &= ~CANWALKUP;
      if(map.playField[j-1][i]->blocker)
        map.playField[j][i]->searchValue &= ~CANWALKLEFT;
      if(map.playField[j+1][i]->blocker)
        map.playField[j][i]->searchValue &= ~CANWALKRIGHT;
      if(map.playField[j][i+1]->blocker)
        map.playField[j][i]->searchValue &= ~CANWALKDOWN;
    }
  }
}

void LoadRunnerPrivate::createObject(KGrObj *o, int x, int y)
{
  map.playField[x][y] = o;
  map.playField[x][y]->init(x, y);
  map.playField[x][y]->move((x-1)*OBJECT_SIZE,(y-1)*OBJECT_SIZE);
  map.playField[x][y]->show();
}

void LoadRunnerPrivate::changeObject(unsigned char kind, int i, int j)
{
  delete map.playField[i][j];
  map.playField[i][j] = 0L;
  
  switch(kind)
  {
    case FREE:
      createObject(new KGrFree(false, view), i, j);
      break;
    case LADDER:
      createObject(new KGrLadder(view), i, j);
      break;
    case HLADDER:
      createObject(new KGrHladder(view), i, j);
      break;
    case BRICK:
      createObject(new KGrBrick(view), i, j);
      break;
    case BETON:
      createObject(new KGrBeton(view), i, j);
      break;
    case FBRICK:
      createObject(new KGrFbrick(view), i, j);
      break;
    case POLE:
      createObject(new KGrPole(view), i, j);
      break;
    case NUGGET:
      createObject(new KGrFree(true, view), i, j);
      nuggets++;
      break;
    case HERO:
      createObject(new KGrFree(false, view), i, j);
      if( hero )
      {
        hero->init(i, j);
        hero->showFigure();
      }
      else
        qDebug("There isn't an hero.");
      break;
    case ENEMY:
    {
      createObject(new KGrFree(false, view), i, j);
      
      KGrEnemy * enemy = new KGrEnemy(i, j);
      
      enemy->setPlayfield( &map );
      enemy->enemyId = enemies.count();
      enemies.append(enemy);
      enemy->showFigure();
      break;
    }
    default :
      createObject(new KGrBrick(view), i, j);
      break;
  } 
}

void LoadRunnerPrivate::deleteLevel()
{
  int X(map.objectsX-1);
  int Y(map.objectsY-1);
  
  for(int i=1; i<Y; i++)
    for(int j=1; j<X; j++)
    {
      delete map.playField[j][i];
      map.playField[j][i] = 0L;
    }
}

void LoadRunnerPrivate::showHidden()
{
  int X(map.objectsX-1);
  int Y(map.objectsY-1);
  
  for(int j=1; j<X; j++)
    for(int i=1; i<Y; i++)
      if( map.playField[j][i]->whatIam() == HLADDER )
        ((KGrHladder *)map.playField[j][i])->showLadder();
  
  initSearchMatrix();
}

void LoadRunnerPrivate::pause()
{
  state = LoadRunner::Pause;
}

void LoadRunnerPrivate::stop()
{
  if( --lifes > 0 )
  {
    loadLevel(level);
    parent_->slotChangeLifes(lifes);
  }
  else
  {
    loadLevel(0);
    state = LoadRunner::Stop;
  }
}

void LoadRunnerPrivate::nextLevel()
{
  level++;
  
  if( level > MAX_LEVELS_NUMBER )
    level = 0;
  
  loadLevel(level);
}

// --------------------------------------------------------
// -------------------- LoadRunner -----------------------
// --------------------------------------------------------

LoadRunner::LoadRunner(QWidget *parent, const char *name)
  : QMainWindow(parent, name)
{
  d = new LoadRunnerPrivate(this);
  
  setCaption( FullAppName );
  
  buildActions();
  
  // build the menu
  QAction * action = 0L;
  
  // build File menu
  QPopupMenu *fileMenu = new QPopupMenu();
  action = d->actions[ACTION_FILE_EXIT];
  if( action )
  {
    action->addTo( fileMenu );
    connect(action, SIGNAL(activated()), this, SLOT(slotFileExit()));
  }
    
  // build Game menu
  QPopupMenu *gameMenu = new QPopupMenu();
  action = d->actions[ACTION_GAME_NEW];
  if( action )
  {
    action->addTo( gameMenu );
    connect(action, SIGNAL(activated()), this, SLOT(slotGameNew()));
  }
    
  action = d->actions[ACTION_GAME_SAVE];
  if( action )
  {
    action->addTo( gameMenu );
    connect(action, SIGNAL(activated()), this, SLOT(slotGameSave()));
  }
    
  action = d->actions[ACTION_GAME_RELOAD];
  if( action )
  {
    action->addTo( gameMenu );
    connect(action, SIGNAL(activated()), this, SLOT(slotGameReload()));
  }
    
  // build Help menu
  QPopupMenu *helpMenu = new QPopupMenu();
  action = d->actions[ACTION_HELP_ABOUT];
  if( action )
  {
    action->addTo( helpMenu );
    connect(action, SIGNAL(activated()), this, SLOT(slotHelpAbout()));
  }
    
  action = d->actions[ACTION_HELP_INDEX];
  if( action )
  {
    action->addTo( helpMenu );
    connect(action, SIGNAL(activated()), this, SLOT(slotHelpIndex()));
  }
    
  menuBar()->insertItem(tr("&File"), fileMenu, 0);
  menuBar()->insertItem(tr("&Game"), gameMenu, 1);
  menuBar()->insertSeparator();
  menuBar()->insertItem(tr("&Help"), helpMenu, -1);
 
  if( d->hero )
  {
    connect(d->hero, SIGNAL(gotNugget()), SLOT(slotIncrementScore()));
    connect(d->hero, SIGNAL(haveAllNuggets()), SLOT(slotShowHidden()));
    connect(d->hero, SIGNAL(caughtHero()), SLOT(slotStop()));
    connect(d->hero, SIGNAL(leaveLevel()), SLOT(slotNextLevel()));
  }
  
//  statusBar()->show();
 
  if( d->view )
    setCentralWidget(d->view);

  srand(1); // initialisiere Random-Generator
}

LoadRunner::~LoadRunner()
{
  // Save config.
  // TODO
  
  menuBar()->clear();
  
  if( d )
  {
    delete d;
    d = 0L;
  }
}

void LoadRunner::buildActions()
{
  // Create all the QAction-s and insert it in d->actions
  QAction * action = 0L;
  
  action = new QAction( tr("Exit from the game."),
                        tr("&Exit"),
#if QT_VERSION < 300
                        0,
#else
                        QKeySequence(),
#endif
                        this );
  d->actions.insert(ACTION_FILE_EXIT, action);
  
  action = new QAction( tr("Start a new game, the last saved game."),
                        QIconLoader::instance()->iconSet("gamenew"),
                        tr("&New Game"),
#if QT_VERSION < 300
                        0,
#else
                        QKeySequence(),
#endif
                        this );
  d->actions.insert(ACTION_GAME_NEW, action);
  
  action = new QAction( tr("Save the name of the game for next time, see 'New Game'."),
                        QIconLoader::instance()->iconSet("gamesave"),
                        tr("&Save Game"),
#if QT_VERSION < 300
                        0,
#else
                        QKeySequence(),
#endif
                        this );
  d->actions.insert(ACTION_GAME_SAVE, action);
  
  action = new QAction( tr("Restart the game from first level."),
                        QIconLoader::instance()->iconSet("gamereload"),
                        tr("&Reload Game"),
#if QT_VERSION < 300
                        0,
#else
                        QKeySequence(),
#endif
                        this );
  d->actions.insert(ACTION_GAME_RELOAD, action);
  
  action = new QAction( tr("Just author and copyright information."),
                        QIconLoader::instance()->iconSet("helpabout"),
                        tr("&About"),
#if QT_VERSION < 300
                        0,
#else
                        QKeySequence(),
#endif
                        this );
  d->actions.insert(ACTION_HELP_ABOUT, action);
  
  action = new QAction( tr("Manual pages."),
                        QIconLoader::instance()->iconSet("helpindex"),
                        tr("&Index"),
#if QT_VERSION < 300
                        0,
#else
                        QKeySequence(),
#endif
                        this );
  d->actions.insert(ACTION_HELP_INDEX, action);
}

void LoadRunner::keyPressEvent(QKeyEvent *k) 
{
  switch( d->state )
  {
    case LoadRunner::NewGame :
      for(QListIterator<KGrEnemy> it(d->enemies); it.current(); ++it)
        it.current()->startSearching(); 
      d->hero->start();
      d->state = LoadRunner::Running;
    case LoadRunner::Pause :
      d->state = LoadRunner::Running;
    case LoadRunner::Running :
      switch(k->key())
      { 
        case Key_Up :
          d->hero->setKey(KGrObj::UP);
          break; 
        case Key_Right :
          d->hero->setKey(KGrObj::RIGHT);
          break; 
        case Key_Down :
          d->hero->setKey(KGrObj::DOWN);
          break; 
        case Key_Left :
          d->hero->setKey(KGrObj::LEFT);
          break; 
        case Key_Control : 
        case Key_Space :
          d->hero->dig();
          break; 
        case Key_K :
          d->stop();
          break;
        case Key_P :
          d->pause();
          break;
        default:
          break;
      }
      break;
    case LoadRunner::Edit :
      break;
    case LoadRunner::Error :
    case LoadRunner::Stop :
    default:
      qDebug("Nothing to do in this state, start a new game.");
      break;
  }
}

void LoadRunner::paintEvent( QPaintEvent * )
{
  switch( d->state )
  {
    case LoadRunner::Error :
    case LoadRunner::Edit :
      break;
    default:
      if( d->hero )
        d->hero->showFigure();
      
      for(EnemiesListIterator it(d->enemies); it.current(); ++it)
        it.current()->showFigure();
      break;
  }
}

void LoadRunner::initStatusBar()
{
//  statusBar->insertItem(klocale->translate("Lifes: 005"),ID_LIFES); 
//  statusBar->insertItem(klocale->translate("Score: 00000"),ID_SCORE);
//  statusBar->insertItem(klocale->translate("Level: 000"),ID_LEVEL);
//  statusBar->insertItem(klocale->translate("Systemlevel"),ID_LEVELPLACE);
//  statusBar->insertItem(" LoadRunner ",ID_DUMMY);
//  statusBar->setAlignment(ID_DUMMY,KStatusBar::Right);
}

void LoadRunner::slotChangeLevel(int l)
{
  QString tmp;
  tmp.setNum(l);
  tmp = tmp.rightJustify(3,'0');
  tmp.insert(0, "Level: ");
//  statusBar->changeItem(tmp,ID_LEVEL);
}

void LoadRunner::slotChangeLifes(int l)
{
  QString tmp;
  tmp.setNum(l);
  tmp = tmp.rightJustify(3,'0');
  tmp.insert(0, "Lifes: ");
//  statusBar->changeItem(tmp,ID_LIFES);
}

void LoadRunner::slotChangeScore(int l)
{
  QString tmp;
  tmp.setNum(l);
  tmp = tmp.rightJustify(5,'0');
  tmp.insert(0, "Score: ");
//  statusBar->changeItem(tmp,ID_SCORE);
}

void LoadRunner::slotIncrementScore()
{
  slotChangeScore(++d->score);
}

void LoadRunner::slotShowHidden()
{
  d->showHidden();
}

void LoadRunner::slotStop()
{
  d->stop();
}

void LoadRunner::slotNextLevel()
{
  d->nextLevel();
  
  slotChangeLevel( d->level );
  slotGameSave();
  
  // change the window caption
  setCaption( QString(AppName) + " - " + d->levelName );
}
  
void LoadRunner::slotFileExit()
{
  qApp->quit();
}

void LoadRunner::slotGameNew()
{
  // read the level from config file
#ifdef _QTE_
  Settings::instance()->setGroup( GAME_GROUP );
  d->level = Settings::instance()->readNumEntry("LEVEL", 0);
#else
  d->level = Settings::instance()->readNumEntry(Settings::instance()->name()
    + "/" + GAME_GROUP + "/LEVEL", 0);
#endif
  
  if( !d->level )
    d->level++;
    
//  qDebug("slotGameNew: Level = %d", d->level);
  
  d->lifes = LIFES_DEFAULT;
  d->score = 0;
  d->nuggets = 0;
  
  d->loadLevel( d->level );
  
  // change the status bar
  slotChangeLevel(d->level);
  slotChangeLifes(d->lifes);
  slotChangeScore(d->score);
  
  // change the window caption
  setCaption( QString(AppName) + " - " + d->levelName );
}

void LoadRunner::slotGameSave()
{
  // write the level to config file
#ifdef _QTE_
  Settings::instance()->setGroup( GAME_GROUP );
  Settings::instance()->writeEntry("LEVEL", d->level);
#else
  Settings::instance()->writeEntry(Settings::instance()->name()+
    "/" + GAME_GROUP + "/LEVEL", d->level);
#endif

  Settings::forceWrite();
}

void LoadRunner::slotGameReload()
{
  d->level = 0;
  
  // write the level to config file
#ifdef _QTE_
  Settings::instance()->setGroup( GAME_GROUP );
  Settings::instance()->writeEntry("LEVEL", d->level);
#else
  Settings::instance()->writeEntry(Settings::instance()->name()+
    "/" + GAME_GROUP + "/LEVEL", d->level);
#endif

  Settings::forceWrite();
  
  // reload the first level
  slotGameNew();
}

void LoadRunner::slotHelpAbout()
{
  About();
}

void LoadRunner::slotHelpIndex()
{
#ifdef _QTE_
  QString helpPath = Dirs::locate( "help/html/eLoadRunner.html" );
  
  if( !helpPath.isEmpty() )
  {
    QCopEnvelope sendShowFile("QPE/Application/helpbrowser", "showFile(QString)");
    sendShowFile << helpPath;
  }
#endif
}

